# Copyright 2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from configparser import ConfigParser

import nacl
import os
import signedjson.key

from sydent.config._base import BaseConfig


class CryptoConfig(BaseConfig):
    def parse_config(self, cfg: "ConfigParser") -> bool:
        """
        Parse the crypto section of the config
        :param cfg: the configuration to be parsed
        """

        signing_key_str = cfg.get("crypto", "ed25519.signingkey")
        signing_key_parts = signing_key_str.split(" ")

        save_key = False

        config_file = os.environ.get("SYDENT_CONF", "sydent.conf")

        if signing_key_str == "":

            if config_file.startswith('/etc/'):

                # Debian's way... (system-wide sydent.conf config file is not writeable)
                print(
                    "INFO: This server does not yet have an ed25519 signing key set. "
                    "I'll create a temporary key for now. Make sure to fix your "
                    f"configuration file '{config_file}' soon by adding an ed25519 "
                    "signing key (run /usr/share/matrix-sydent/generate-key "
                    "to create one). On Debian, an ed25519 signing key should "
                    "normally have been created during package installation. If not, "
                    "please report it as a bug against 'matrix-sydent' in "
                    "Debian's bug tracker. Thanks!"
                )

            else:

                # upstream's way...
                print(
                    "INFO: This server does not yet have an ed25519 signing key. "
                    "Creating one and saving it in the config file."
                )

            self.signing_key = signedjson.key.generate_signing_key("0")

            save_key = True
        elif len(signing_key_parts) == 1:
            # old format key
            print("INFO: Updating signing key format: brace yourselves")

            self.signing_key = nacl.signing.SigningKey(
                signing_key_str, encoder=nacl.encoding.HexEncoder
            )
            self.signing_key.version = "0"
            self.signing_key.alg = signedjson.key.NACL_ED25519

            save_key = True
        else:
            self.signing_key = signedjson.key.decode_signing_key_base64(
                signing_key_parts[0], signing_key_parts[1], signing_key_parts[2]
            )

        if save_key:
            signing_key_str = "%s %s %s" % (
                self.signing_key.alg,
                self.signing_key.version,
                signedjson.key.encode_signing_key_base64(self.signing_key),
            )

            # only save the config, if a non-system-wide config file is in use
            if not config_file.startswith('/etc/'):
                cfg.set("crypto", "ed25519.signingkey", signing_key_str)
                return True
            else:
                return False

        else:
            return False
